#include "usql.h"
#include "exception.h"
#include "csvreader.h"
#include "ml_date.h"

#include <algorithm>
#include <fstream>

namespace usql {

std::unique_ptr<Table> USql::execute(const std::string &command) {
	try {
		std::unique_ptr<Node> node = m_parser.parse(command);
		return execute(*node);

	} catch (std::exception &e) {
		return create_stmt_result_table(-1, e.what());
	}

}

std::unique_ptr<Table> USql::execute(Node &node) {
	// TODO optimize execution nodes here
	switch (node.node_type) {
		case NodeType::create_table:
			return execute_create_table(static_cast<CreateTableNode &>(node));
		case NodeType::create_table_as_select:
			return execute_create_table_as_table(static_cast<CreateTableAsSelectNode &>(node));
		case NodeType::insert_into:
			return execute_insert_into_table(static_cast<InsertIntoTableNode &>(node));
		case NodeType::select_from:
			return execute_select(static_cast<SelectFromTableNode &>(node));
		case NodeType::delete_from:
			return execute_delete(static_cast<DeleteFromTableNode &>(node));
		case NodeType::update_table:
			return execute_update(static_cast<UpdateTableNode &>(node));
		case NodeType::load_table:
			return execute_load(static_cast<LoadIntoTableNode &>(node));
		case NodeType::save_table:
			return execute_save(static_cast<SaveTableNode &>(node));
		default:
			return create_stmt_result_table(-1, "unknown statement");
	}
}


std::unique_ptr<Table> USql::execute_create_table(CreateTableNode &node) {
	// TODO check table does not exists
	Table table{node.table_name, node.cols_defs};
	m_tables.push_back(table);

	return create_stmt_result_table(0, "table created");
}


std::unique_ptr<Table> USql::execute_create_table_as_table(CreateTableAsSelectNode &node) {
	// TODO check table does not exists

	auto select = execute_select((SelectFromTableNode &) *node.select_table);

	// create table
	Table new_table{node.table_name, select->m_col_defs};
	m_tables.push_back(new_table);

	// copy rows
	// must be here, if rows are put into new_table, they are lost during m_tables.push_table
	Table *table = find_table(node.table_name);
	for( Row& orig_row : select->m_rows) {
		table->addCopyOfRow(orig_row);
	}

	select.release();	// is it correct? hoping not to release select table here and then when releasing CreateTableAsSelectNode

	return create_stmt_result_table(0, "table created");
}


std::unique_ptr<Table> USql::execute_insert_into_table(InsertIntoTableNode &node) {
	// TODO check column names.size = values.size

	// find table
	Table *table_def = find_table(node.table_name);

	// prepare empty new_row
	Row new_row = table_def->createEmptyRow();

	// copy values
	for (size_t i = 0; i < node.cols_names.size(); i++) {
		ColDefNode col_def = table_def->get_column_def(node.cols_names[i].name);
		auto col_value = evalValueNode(table_def, new_row, node.cols_values[i].get());

		new_row.setColumnValue(&col_def, col_value.get());
	}

	// append new_row
	table_def->addRow(new_row);

	return create_stmt_result_table(0, "insert succeeded");
}


std::unique_ptr<Table> USql::execute_select(SelectFromTableNode &node) {
	// find source table
	Table *table = find_table(node.table_name);


	// create result table
	std::vector<ColDefNode> result_tbl_col_defs{};
	std::vector<int> source_table_col_index{};

	for (int i = 0; i < node.cols_names->size(); i++) {
		auto [ src_tbl_col_index, rst_tbl_col_def ]  = getColumnDefinition(table, &node.cols_names->operator[](i), i);

		source_table_col_index.push_back(src_tbl_col_index);
		result_tbl_col_defs.push_back(rst_tbl_col_def);
	}
	auto result = std::make_unique<Table>("result", result_tbl_col_defs);


	// execute access plan
	for (auto row = begin(table->m_rows); row != end(table->m_rows); ++row) {
		// eval where for row
		if (evalWhere(node.where.get(), table, *row)) {
			// prepare empty row
			Row new_row = result->createEmptyRow();

			// copy column values
			for (auto idx = 0; idx < result->columns_count(); idx++) {
				auto row_col_index = source_table_col_index[idx];

				if (row_col_index == -1) {	// TODO introduce constant here
					auto evaluated_value = evalValueNode(table, *row, node.cols_names->operator[](idx).value.get());
					ValueNode *col_value = evaluated_value.get();

					new_row.setColumnValue(&result_tbl_col_defs[idx], col_value);
				} else {
					ColValue *col_value = row->ithColumn(row_col_index);
					new_row.setColumnValue(&result_tbl_col_defs[idx], col_value);
				}
			}

			// add row to result
			result->m_rows.push_back(new_row);
		}
	}

	return std::move(result);
}

std::tuple<int, ColDefNode> USql::getColumnDefinition(Table *table, SelectColNode *select_col_node, int col_order ) {
	std::string new_col_name = select_col_node->name;

	if (select_col_node->value->node_type == NodeType::column_name) {
		ColDefNode src_cdef = table->get_column_def(new_col_name);
		ColDefNode cdef = ColDefNode{new_col_name, src_cdef.type, col_order, src_cdef.length, src_cdef.null};
		return std::make_tuple(src_cdef.order, cdef);

	} else if (select_col_node->value->node_type == NodeType::function) {
		auto node = static_cast<FunctionNode *>(select_col_node->value.get());

		if (node->function == "to_string") {
			ColDefNode cdef = ColDefNode{new_col_name, ColumnType::varchar_type, col_order, 64, true};
			return std::make_tuple(-1, cdef);
		} else if (node->function == "to_date") {
			ColDefNode cdef = ColDefNode{new_col_name, ColumnType::integer_type, col_order, 1, true};
			return std::make_tuple(-1, cdef);
		}
		throw Exception("Unsupported function");
	}
	throw Exception("Unsupported node type");
}


std::unique_ptr<Table> USql::execute_delete(DeleteFromTableNode &node) {
	// find source table
	Table *table = find_table(node.table_name);

	// execute access plan
	auto it = table->m_rows.begin();
	for (; it != table->m_rows.end();) {
		if (evalWhere(node.where.get(), table, *it)) {
			// TODO this can be really expensive operation
			it = table->m_rows.erase(it);
		} else {
			++it;
		}
	}

	return create_stmt_result_table(0, "delete succeeded");
}


std::unique_ptr<Table> USql::execute_update(UpdateTableNode &node) {
	// find source table
	Table *table = find_table(node.table_name);

	// execute access plan
	for (auto row = begin(table->m_rows); row != end(table->m_rows); ++row) {
		// eval where for row
		if (evalWhere(node.where.get(), table, *row)) {
			int i = 0;
			for (const auto& col : node.cols_names) {
				ColDefNode col_def = table->get_column_def(col.name);	// TODO cache it like in select
				std::unique_ptr<ValueNode> new_val = evalArithmeticOperator(col_def.type,
						      static_cast<ArithmeticalOperatorNode &>(*node.values[i]), table, *row);

				row->setColumnValue(&col_def, new_val.get());
				i++;
			}
		}
	}

	return create_stmt_result_table(0, "delete succeeded");
}


std::unique_ptr<Table> USql::execute_load(LoadIntoTableNode &node) {
	// find source table
	Table *table_def = find_table(node.table_name);

	// read data
	std::ifstream ifs(node.filename);
	std::string content((std::istreambuf_iterator<char>(ifs)),
			    (std::istreambuf_iterator<char>()));

	CsvReader csvparser{};
	auto csv = csvparser.parseCSV(content);

	std::vector<ColDefNode> &colDefs = table_def->m_col_defs;

	for (auto it = csv.begin() + 1; it != csv.end(); ++it) {
		std::vector<std::string> csv_line = *it;

		// prepare empty new_row
		Row new_row = table_def->createEmptyRow();

		// copy values
		for (size_t i = 0; i < table_def->columns_count(); i++) {
			ColDefNode col_def = table_def->get_column_def(colDefs[i].name);

			// TODO validate value
			if (col_def.type == ColumnType::integer_type) {
				new_row.setColumnValue(col_def.order, std::stol(csv_line[i]));
			} else if (col_def.type == ColumnType::float_type) {
				new_row.setColumnValue(col_def.order, std::stof(csv_line[i]));
			} else {
				new_row.setColumnValue(col_def.order, csv_line[i]);
			}
		}

		// append new_row
		table_def->addRow(new_row);
	}

	return create_stmt_result_table(0, "load succeeded");
}


std::unique_ptr<Table> USql::execute_save(SaveTableNode &node) {
	// find source table
	Table *table_def = find_table(node.table_name);

	// header
	std::string out_string;
	for(int i = 0; i < table_def->m_col_defs.size(); i++) {
		if (i > 0) out_string += ",";
		out_string += table_def->m_col_defs[i].name;
	}

	// rows
	for (auto it = table_def->m_rows.begin() + 1; it != table_def->m_rows.end(); ++it) {
		std::string csv_line;
		for(int i = 0; i < table_def->m_col_defs.size(); i++) {
			if (i > 0) csv_line += ",";

			auto col = it->ithColumn(i);
			if (!col->isNull()) {
				csv_line += col->getStringValue();		// TODO handle enclosing commas etc
			}
		}
		out_string += "\n";
		out_string += csv_line;
	}

	// save data
	std::ofstream file(node.filename);
	file << out_string;
	file.close();

	return create_stmt_result_table(0, "save succeeded");
}


bool USql::evalWhere(Node *where, Table *table, Row &row) const {
	switch (where->node_type) {    // no where clause
		case NodeType::true_node:
			return true;
		case NodeType::relational_operator:     //  just one condition
			return evalRelationalOperator(*((RelationalOperatorNode *) where), table, row);
		case NodeType::logical_operator:
			return evalLogicalOperator(*((LogicalOperatorNode *) where), table, row);
		default:
			throw Exception("Wrong node type");
	}

	return false;
}


bool USql::evalRelationalOperator(const RelationalOperatorNode &filter, Table *table, Row &row) const {
	std::unique_ptr<ValueNode> left_value = evalValueNode(table, row, filter.left.get());
	std::unique_ptr<ValueNode> right_value = evalValueNode(table, row, filter.right.get());

	double comparator;

	if (left_value->node_type == NodeType::int_value && right_value->node_type == NodeType::int_value) {
		comparator = left_value->getIntValue() - right_value->getIntValue();
	} else if ((left_value->node_type == NodeType::int_value && right_value->node_type == NodeType::float_value) ||
		(left_value->node_type == NodeType::float_value && right_value->node_type == NodeType::int_value) ||
		(left_value->node_type == NodeType::float_value && right_value->node_type == NodeType::float_value)) {
		comparator = left_value->getDoubleValue() - right_value->getDoubleValue();
	} else if (left_value->node_type == NodeType::string_value || right_value->node_type == NodeType::string_value) {
		comparator = left_value->getStringValue().compare(right_value->getStringValue());
	} else {
		// TODO throw exception
	}

	switch (filter.op) {
		case RelationalOperatorType::equal:
			return comparator == 0.0;
		case RelationalOperatorType::not_equal:
			return comparator != 0.0;
		case RelationalOperatorType::greater:
			return comparator > 0.0;
		case RelationalOperatorType::greater_equal:
			return comparator >= 0.0;
		case RelationalOperatorType::lesser:
			return comparator < 0.0;
		case RelationalOperatorType::lesser_equal:
			return comparator <= 0.0;
	}

	throw Exception("invalid relational operator");
}


std::unique_ptr<ValueNode> USql::evalValueNode(Table *table, Row &row, Node *node) {
	if (node->node_type == NodeType::database_value || node->node_type == NodeType::column_name) {	// TODO sjednotit
		return evalDatabaseValueNode(table, row, node);

	} else if (node->node_type == NodeType::int_value || node->node_type == NodeType::float_value || node->node_type == NodeType::string_value) {
		return evalLiteralValueNode(table, row, node);

	} else if (node->node_type == NodeType::function) {
		return evalFunctionValueNode(table, row, node);
	}
	throw Exception("unsupported node type");
}


std::unique_ptr<ValueNode> USql::evalDatabaseValueNode(Table *table, Row &row, Node *node) {
	auto *dvl = static_cast<DatabaseValueNode *>(node);
	ColDefNode col_def = table->get_column_def( dvl->col_name); // TODO optimize it to just get this def once
	auto db_value = row.ithColumn(col_def.order);

	if (col_def.type == ColumnType::integer_type) {
		return std::make_unique<IntValueNode>(db_value->getIntValue());
	}
	if (col_def.type == ColumnType::float_type) {
		return std::make_unique<DoubleValueNode>(db_value->getDoubleValue());
	}
	if (col_def.type == ColumnType::varchar_type) {
		return std::make_unique<StringValueNode>(db_value->getStringValue());
	}
	throw Exception("unknown database value type");
}


std::unique_ptr<ValueNode> USql::evalLiteralValueNode(Table *table, Row &row, Node *node) {
	if (node->node_type == NodeType::int_value) {
		auto *ivl = static_cast<IntValueNode *>(node);
		return std::make_unique<IntValueNode>(ivl->value);

	} else if (node->node_type == NodeType::float_value) {
		auto *ivl = static_cast<DoubleValueNode *>(node);
		return std::make_unique<DoubleValueNode>(ivl->value);

	} else if (node->node_type == NodeType::string_value) {
		auto *ivl = static_cast<StringValueNode *>(node);
		return std::make_unique<StringValueNode>(ivl->value);

	}

	throw Exception("invalid type");
}


std::unique_ptr<ValueNode> USql::evalFunctionValueNode(Table *table, Row &row, Node *node) {
	auto  *fnc = static_cast<FunctionNode *>(node);

	std::vector<std::unique_ptr<ValueNode>> evaluatedPars;
	for(auto & param : fnc->params) {
		evaluatedPars.push_back(evalValueNode(table, row, param.get()));
	}

	// TODO use some enum
	if (fnc->function == "lower") {
		std::string str = evaluatedPars[0]->getStringValue();
		std::transform(str.begin(), str.end(), str.begin(), [](unsigned char c) -> unsigned char { return std::tolower(c); });
		return std::make_unique<StringValueNode>(str);
	}
	if (fnc->function == "upper") {
		std::string str = evaluatedPars[0]->getStringValue();
		std::transform(str.begin(), str.end(), str.begin(), [](unsigned char c) -> unsigned char { return std::toupper(c); });
		return std::make_unique<StringValueNode>(str);
	}

	if (fnc->function == "to_date") {
		std::string date = evaluatedPars[0]->getStringValue();
		std::string format = evaluatedPars[1]->getStringValue();
		long epoch_time = string_to_date(date, format);
		return std::make_unique<IntValueNode>(epoch_time);
	}
	if (fnc->function == "to_string") {
		long date = evaluatedPars[0]->getIntValue();
		std::string format = evaluatedPars[1]->getStringValue();
		std::string formated_date = date_to_string(date, format);
		return std::make_unique<StringValueNode>(formated_date);
	}

	throw Exception("invalid function");
}


bool USql::evalLogicalOperator(LogicalOperatorNode &node, Table *pTable, Row &row) const {
	bool left = evalRelationalOperator(static_cast<const RelationalOperatorNode &>(*node.left), pTable, row);

	if ((node.op == LogicalOperatorType::and_operator && !left) || (node.op == LogicalOperatorType::or_operator && left))
		return left;

	bool right = evalRelationalOperator(static_cast<const RelationalOperatorNode &>(*node.right), pTable, row);
	return right;
}


std::unique_ptr<ValueNode> USql::evalArithmeticOperator(ColumnType outType, ArithmeticalOperatorNode &node, Table *table, Row &row) const {
	if (node.op == ArithmeticalOperatorType::copy_value) {
		return evalValueNode(table, row, node.left.get());
	}

	std::unique_ptr<ValueNode> left = evalValueNode(table, row, node.left.get());
	std::unique_ptr<ValueNode> right = evalValueNode(table, row, node.right.get());

	if (outType == ColumnType::float_type) {
		double l = ((ValueNode *) left.get())->getDoubleValue();
		double r = ((ValueNode *) right.get())->getDoubleValue();
		switch (node.op) {
			case ArithmeticalOperatorType::plus_operator:
				return std::make_unique<DoubleValueNode>(l + r);
			case ArithmeticalOperatorType::minus_operator:
				return std::make_unique<DoubleValueNode>(l - r);
			case ArithmeticalOperatorType::multiply_operator:
				return std::make_unique<DoubleValueNode>(l * r);
			case ArithmeticalOperatorType::divide_operator:
				return std::make_unique<DoubleValueNode>(l / r);
			default:
				throw Exception("implement me!!");
		}
	} else if (outType == ColumnType::integer_type) {
		long l = ((ValueNode *) left.get())->getIntValue();
		long r = ((ValueNode *) right.get())->getIntValue();
		switch (node.op) {
			case ArithmeticalOperatorType::plus_operator:
				return std::make_unique<IntValueNode>(l + r);
			case ArithmeticalOperatorType::minus_operator:
				return std::make_unique<IntValueNode>(l - r);
			case ArithmeticalOperatorType::multiply_operator:
				return std::make_unique<IntValueNode>(l * r);
			case ArithmeticalOperatorType::divide_operator:
				return std::make_unique<IntValueNode>(l / r);
			default:
				throw Exception("implement me!!");
		}

	} else if (outType == ColumnType::varchar_type) {
		std::string l = ((ValueNode *) left.get())->getStringValue();
		std::string r = ((ValueNode *) right.get())->getStringValue();
		switch (node.op) {
			case ArithmeticalOperatorType::plus_operator:
				return std::make_unique<StringValueNode>(l + r);
			default:
				throw Exception("implement me!!");
		}
	}

	throw Exception("implement me!!");
}


std::unique_ptr<Table> USql::create_stmt_result_table(long code, const std::string& text) {
	std::vector<ColDefNode> result_tbl_col_defs{};
	result_tbl_col_defs.push_back(ColDefNode("code", ColumnType::integer_type, 0, 1, false));
	result_tbl_col_defs.push_back(ColDefNode("desc", ColumnType::varchar_type, 1, 255, false));

	auto table_def = std::make_unique<Table>("result", result_tbl_col_defs);

	Row new_row = table_def->createEmptyRow();
	new_row.setColumnValue(0, code);
	new_row.setColumnValue(1, text);
	table_def->addRow(new_row);

	return std::move(table_def);
}



Table *USql::find_table(const std::string &name) {
	auto name_cmp = [name](const Table& t) { return t.m_name == name; };
	auto table_def = std::find_if(begin(m_tables), end(m_tables), name_cmp);
	if (table_def != std::end(m_tables)) {
		return table_def.operator->();
	} else {
		throw Exception("table not found (" + name + ")");
	}
}

}	// namespace